import dataclasses
from functools import cached_property
from typing import ClassVar


@dataclasses.dataclass
class CppNamedType:
    type_name: str
    namespace: ClassVar[str] = ""

    @cached_property
    def name(self) -> str:
        return self.namespace + self.type_name


class QtGqlNamedType(CppNamedType):
    namespace = "qtgql::"


class QtGqlTypes:
    QGraphQLList = QtGqlNamedType("QGraphQLList")



from functools import cached_property
from pathlib import Path
from typing import Callable
from typing import Type

from attrs import define

from qtgqlcodegen.compiler.template import schema_types_template
from qtgqlcodegen.compiler.template import SchemaTemplateContext
from qtgqlcodegen.introspection import SchemaEvaluator
from qtgqlcodegen.runtime.custom_scalars import CUSTOM_SCALARS
from qtgqlcodegen.runtime.custom_scalars import CustomScalarMap


@define(slots=False)
class QtGqlConfig:
    """Encapsulates configurations for a qtgql-codegen application per GraphQL
    schema."""

    graphql_dir: Path
    """A directory contains.

    - schema.graphql, represents the current schema definition at the server.
    - operations.graphql, queries, mutations and subscription handlers would be generated based on the operations defined there.
    """
    env_name: str = "QGqlEnv"
    """The generated types would find the environment by this name.

    Also the generated QML imports would fall under this namespace.
    """
    evaluator: Type[SchemaEvaluator] = SchemaEvaluator
    """evaluates the schema and generates types."""
    custom_scalars: CustomScalarMap = {}
    """mapping of custom scalars, respected by the schema evaluator."""
    template_class: Callable[[SchemaTemplateContext], str] = schema_types_template
    """jinja template."""

    @cached_property
    def schema_path(self) -> Path:
        return self.graphql_dir / "schema.graphql"

    @cached_property
    def operations_dir(self) -> Path:
        return self.graphql_dir / "operations.graphql"

    @cached_property
    def generated_dir(self):
        ret = self.graphql_dir / "__generated__"
        if not ret.exists():
            ret.mkdir()
        return ret

    @cached_property
    def _evaluator(self) -> SchemaEvaluator:
        return self.evaluator(self)

    def generate(self) -> None:
        self._evaluator.dump()

    def __attrs_post_init__(self):
        if self.custom_scalars != CUSTOM_SCALARS:
            self.custom_scalars.update(CUSTOM_SCALARS)



from __future__ import annotations

from collections import defaultdict
from functools import cached_property
from typing import Optional
from typing import TYPE_CHECKING

import attrs
import graphql
from attr import define
from frozendict import frozendict
from typingref import UNSET

from qtgqlcodegen.compiler.template import config_template
from qtgqlcodegen.compiler.template import ConfigContext
from qtgqlcodegen.graphql_ref import has_id_selection
from qtgqlcodegen.graphql_ref import has_typename_selection
from qtgqlcodegen.graphql_ref import inject_id_selection
from qtgqlcodegen.graphql_ref import inject_typename_selection
from qtgqlcodegen.graphql_ref import is_field_node
from qtgqlcodegen.graphql_ref import is_inline_fragment

if TYPE_CHECKING:
    from graphql import language as gql_lang
    from graphql.type import definition as gql_def

    from qtgqlcodegen.introspection import SchemaEvaluator
    from qtgqlcodegen.objecttype import QtGqlVariableDefinition, GqlTypeHinter
    from qtgqlcodegen.objecttype import QtGqlFieldDefinition
    from qtgqlcodegen.objecttype import QtGqlObjectTypeDefinition


def is_type_name_selection(field_node: gql_lang.FieldNode):
    # typename is not a 'real' selection and is handled with special care.
    if field_node.name.value == "__typename":
        return True
    return False


def get_field_from_field_node(
    selection: gql_lang.FieldNode,
    field_type: QtGqlObjectTypeDefinition,
    operation: QtGqlOperationDefinition,
    parent_interface_field: Optional[QtGqlQueriedField] = UNSET,
) -> QtGqlQueriedField:
    field_node = is_field_node(selection)
    assert field_node
    return QtGqlQueriedField.from_field(
        field_type.fields_dict[field_node.name.value],
        selection_set=field_node.selection_set,
        operation=operation,
        parent_interface_field=parent_interface_field,
    )


@attrs.define(frozen=True, slots=False)
class QtGqlQueriedField:
    definition: QtGqlFieldDefinition
    operation: QtGqlOperationDefinition
    choices: frozendict[str, dict[str, QtGqlQueriedField]] = attrs.Factory(frozendict)
    selections: dict[str, QtGqlQueriedField] = attrs.Factory(dict)
    narrowed_type: Optional[QtGqlQueriedObjectType] = None

    def __repr__(self):
        return (
            f"{self.definition.name}[{', '.join(self.selections.keys())}\n"
            + "...on".join([f"{k} {repr(v)}" for k, v in self.choices.items()])
            + "]"
        )

    @cached_property
    def type(self) -> GqlTypeHinter:
        return self.definition.type

    @cached_property
    def property_type(self) -> str:
        if self.definition.type.is_object_type:
            assert self.narrowed_type
            return self.narrowed_type.name
        if cs := self.definition.is_custom_scalar:
            return cs.property_type
        return self.type.member_type

    @property
    def proxy_of(self) -> GqlTypeHinter:
        ret = self.definition.type
        assert ret
        return ret

    @property
    def name(self) -> str:
        return self.definition.name

    @classmethod
    def from_field(
        cls,
        field_definition: QtGqlFieldDefinition,
        selection_set: Optional[gql_lang.SelectionSetNode],
        operation: QtGqlOperationDefinition,
        parent_interface_field: Optional[QtGqlQueriedField] = UNSET,
        is_root: bool = False,
    ) -> QtGqlQueriedField:
        """Main purpose here is to find inner selections of fields, this could
        be an object type, interface, union or a list.

        Any other fields should not have inner selections.
        """
        assert parent_interface_field is not UNSET
        if not hasattr(selection_set, "selections"):
            return cls(definition=field_definition, operation=operation)
        assert selection_set
        tp = field_definition.type
        if (
            tp.is_model
        ):  # GraphQL's lists are basically the object beneath them in terms of selections.
            tp = tp.is_model

        tp_is_union = tp.is_union

        # inject id selection for types that supports it. unions are handled below.
        if field_definition.can_select_id and not has_id_selection(selection_set):
            inject_id_selection(selection_set)

        selections: dict[str, QtGqlQueriedField] = {}
        choices: defaultdict[str, dict[str, QtGqlQueriedField]] = defaultdict(dict)
        narrowed_type: Optional[QtGqlQueriedObjectType] = None
        # inject parent interface selections.
        if (tp.is_object_type or tp.is_interface) and parent_interface_field:
            selections.update({f.name: f for f in parent_interface_field.selections.values()})

        if tp_is_union:
            for selection in selection_set.selections:
                fragment = is_inline_fragment(selection)
                assert fragment

                type_name = fragment.type_condition.name.value
                concrete = operation.evaluator.get_objecttype_by_name(type_name)
                assert concrete
                if not has_typename_selection(fragment.selection_set):
                    inject_typename_selection(fragment.selection_set)
                if not has_id_selection(fragment.selection_set) and concrete.has_id_field:
                    inject_id_selection(fragment.selection_set)

                for selection_node in fragment.selection_set.selections:
                    field_node = is_field_node(selection_node)
                    assert field_node

                    if not is_type_name_selection(field_node):
                        __f = get_field_from_field_node(
                            field_node,
                            concrete,
                            operation,
                            parent_interface_field,
                        )
                        choices[type_name][field_definition.name] = __f

        elif interface_def := tp.is_interface:
            # first get all linear selections.
            for selection in selection_set.selections:
                if not is_inline_fragment(selection):
                    field_node = is_field_node(selection)
                    assert field_node
                    if not is_type_name_selection(field_node):
                        __f = get_field_from_field_node(
                            field_node,
                            interface_def,
                            operation,
                            parent_interface_field,
                        )
                        selections[__f.name] = __f

            for selection in selection_set.selections:
                if inline_frag := is_inline_fragment(selection):
                    type_name = inline_frag.type_condition.name.value
                    # no need to validate inner types are implementation, graphql-core does this.
                    concrete = operation.evaluator.get_objecttype_by_name(
                        type_name,
                    ) or operation.evaluator.get_interface_by_name(type_name)
                    assert concrete
                    for inner_selection in inline_frag.selection_set.selections:
                        field_node = is_field_node(inner_selection)
                        assert field_node
                        if not is_type_name_selection(field_node):
                            __f = get_field_from_field_node(
                                field_node,
                                concrete,
                                operation,
                                parent_interface_field,
                            )
                            choices[type_name][field_definition.name] = __f

        else:  # object types.
            obj_def = tp.is_object_type
            assert obj_def
            for selection in selection_set.selections:
                field_node = is_field_node(selection)
                assert field_node
                if not is_type_name_selection(field_node):
                    __f = get_field_from_field_node(
                        field_node,
                        obj_def,
                        operation,
                        parent_interface_field,
                    )
                    selections[__f.name] = __f
            queried_obj = QtGqlQueriedObjectType(
                definition=obj_def,
                fields=selections,
            )
            operation.narrowed_types_map[queried_obj.name] = queried_obj
            narrowed_type = queried_obj

        def sorted_distinct_fields(
            fields: dict[str, QtGqlQueriedField],
        ) -> dict[str, QtGqlQueriedField]:
            return dict(sorted(fields.items()))

        return cls(
            definition=field_definition,
            selections=sorted_distinct_fields(selections),
            choices=frozendict({k: sorted_distinct_fields(v) for k, v in choices.items()}),
            operation=operation,
            narrowed_type=narrowed_type,
        )

    def as_conf_string(self) -> str:
        return config_template(
            context=ConfigContext(self),
        )


@define(slots=False)
class QtGqlQueriedObjectType:
    definition: QtGqlObjectTypeDefinition = attrs.field(on_setattr=attrs.setters.frozen)
    fields: dict[str, QtGqlQueriedField] = attrs.Factory(dict)
    is_root_field: bool = False

    @cached_property
    def name(self) -> str:
        return f"{self.definition.name}__{'$'.join(sorted(self.fields.keys()))}"

    @cached_property
    def doc_fields(self) -> str:
        return "{} {{\n  {}\n}}".format(
            self.definition.name,
            "\n   ".join(self.fields.keys()),
        )

    @cached_property
    def references(self) -> list[QtGqlQueriedField]:
        return [f for f in self.fields.values() if f.type.is_object_type]


@define(slots=False)
class QtGqlOperationDefinition:
    operation_def: gql_def.OperationDefinitionNode
    evaluator: SchemaEvaluator
    directives: list[str] = attrs.Factory(list)
    fragments: list[str] = attrs.Factory(list)
    variables: list[QtGqlVariableDefinition] = attrs.Factory(list)
    narrowed_types_map: dict[str, QtGqlQueriedObjectType] = attrs.Factory(dict)

    def __attrs_post_init__(self) -> None:
        # instantiating the queried fields here, they build the narrowed types.
        self.root_field  # noqa

    @property
    def operation_config(self) -> str:
        return self.root_field.as_conf_string()

    @property
    def name(self) -> str:
        assert self.operation_def.name
        return self.operation_def.name.value

    @cached_property
    def query(self) -> str:
        return graphql.print_ast(self.operation_def)

    @cached_property
    def narrowed_types(self) -> tuple[QtGqlQueriedObjectType, ...]:
        return tuple(self.narrowed_types_map.values())

    @cached_property
    def _root_type(self) -> QtGqlObjectTypeDefinition:
        root_type = self.evaluator.operation_types.get(self.operation_def.operation, None)
        assert root_type, f"You don't have a {self.operation_def.operation} type on your schema"
        return root_type

    @cached_property
    def root_field(self) -> QtGqlQueriedField:
        root_field_def: gql_lang.FieldNode = self.operation_def.selection_set.selections[0]  # type: ignore
        return QtGqlQueriedField.from_field(
            self._root_type.fields_dict[root_field_def.name.value],
            root_field_def.selection_set,
            self,
            parent_interface_field=None,
            is_root=True,
        )

    @classmethod
    def from_definition(
        cls,
        operation_def: gql_def.OperationDefinitionNode,
        evaluator: SchemaEvaluator,
        directives: list[str],
        variables: list[QtGqlVariableDefinition],
    ):
        return cls(
            operation_def=operation_def,
            evaluator=evaluator,
            directives=directives,
            variables=variables,
        )



from functools import cached_property
from typing import NewType
from typing import Optional

from attr import define

CType = NewType("CType", str)


@define(slots=False)
class BuiltinScalar:
    tp: CType
    defaults_member_name: str
    graphql_name: str
    from_json_convertor: str

    @cached_property
    def default_value(self) -> str:
        return f"qtgql::bases::DEFAULTS::{self.defaults_member_name}"


class _BuiltinScalars:
    INT = BuiltinScalar(
        CType("int"),
        "INT",
        graphql_name="Int",
        from_json_convertor="toInt()",
    )
    FLOAT = BuiltinScalar(
        CType("float"),
        "FLOAT",
        graphql_name="Float",
        from_json_convertor="toDouble()",
    )
    STRING = BuiltinScalar(
        CType("QString"),
        "STRING",
        graphql_name="String",
        from_json_convertor="toString()",
    )
    ID = BuiltinScalar(
        CType("QString"),
        "ID",
        graphql_name="ID",
        from_json_convertor="toString()",
    )
    BOOLEAN = BuiltinScalar(
        CType("bool"),
        "BOOL",
        graphql_name="Boolean",
        from_json_convertor="toBool()",
    )
    UUID = BuiltinScalar(
        CType("QUuid"),
        "UUID",
        graphql_name="UUID",
        from_json_convertor="toVariant().toUuid()",
    )

    @classmethod
    def __iter__(cls):
        for name, member in cls.__dict__.items():
            if name.isupper():
                yield member

    def by_graphql_name(self, name: str) -> Optional[BuiltinScalar]:
        for scalar in self:
            if scalar.graphql_name == name:
                return scalar

    def keys(self) -> list[str]:
        return [scalar.graphql_name for scalar in self]


BuiltinScalars = _BuiltinScalars()



from __future__ import annotations

from typing import TYPE_CHECKING

from attrs import define
from jinja2 import Environment
from jinja2 import PackageLoader
from jinja2 import select_autoescape

if TYPE_CHECKING:  # pragma: no cover
    from qtgqlcodegen.config import QtGqlConfig
    from qtgqlcodegen.compiler.operation import QtGqlOperationDefinition, QtGqlQueriedField
    from qtgqlcodegen.utils import FileSpec
    from qtgqlcodegen.objecttype import (
        QtGqlEnumDefinition,
        QtGqlInputObjectTypeDefinition,
        QtGqlInterfaceDefinition,
        QtGqlObjectTypeDefinition,
    )


template_env = Environment(
    loader=PackageLoader("qtgqlcodegen"),
    autoescape=select_autoescape(),
    variable_start_string="👉",  # originally {{ variable }}, using 👉 variable 👈 because C++ uses curly brackets.
    variable_end_string="👈",
)


def wrap_curly_filter(v: str, ignore: bool = False) -> str:
    if ignore:
        return v
    return "{" + v + "}"


template_env.filters["wrapcurly"] = wrap_curly_filter

SCHEMA_TEMPLATE = template_env.get_template("schema.jinja.hpp")
OPERATION_TEMPLATE = template_env.get_template("operation.jinja.hpp")
CONFIG_TEMPLATE = template_env.get_template("config.jinja.hpp")
CMAKE_TEMPLATE = template_env.get_template("CMakeLists.jinja.cmake")


@define
class SchemaTemplateContext:
    enums: list[QtGqlEnumDefinition]
    types: list[QtGqlObjectTypeDefinition]
    interfaces: list[QtGqlInterfaceDefinition]
    input_objects: list[QtGqlInputObjectTypeDefinition]
    config: QtGqlConfig

    @property
    def dependencies(self) -> list[str]:
        ret: list[str] = []
        for scalar in self.config.custom_scalars.values():
            if scalar.include_path not in ret:
                ret.append(scalar.include_path)
        return [f"#include {inc}" for inc in ret]

    @property
    def custom_scalars(self) -> list[str]:
        return [scalar.graphql_name for scalar in self.config.custom_scalars.values()]


@define(slots=False)
class OperationTemplateContext:
    operation: QtGqlOperationDefinition
    config: QtGqlConfig

    @property
    def ns(self) -> str:
        return self.operation.name.lower()

    @property
    def schema_ns(self) -> str:
        return self.config.env_name


def schema_types_template(context: SchemaTemplateContext) -> str:
    return SCHEMA_TEMPLATE.render(context=context)


def operation_template(context: OperationTemplateContext) -> str:
    return OPERATION_TEMPLATE.render(context=context)


@define(slots=False)
class CmakeTemplateContext:
    config: QtGqlConfig
    sources: list[FileSpec]

    @property
    def target_name(self) -> str:
        return self.config.env_name


def cmake_template(context: SchemaTemplateContext) -> str:
    return CMAKE_TEMPLATE.render(context=context)


@define
class ConfigContext:
    p_field: QtGqlQueriedField

    @property
    def choices(self):
        if self.p_field.choices:
            return {
                type_name: {
                    selection.name: selection.as_conf_string() or "None" for selection in selections
                }
                for type_name, selections in self.p_field.choices.items()
            }
        else:
            return {}

    @property
    def selections(self) -> dict[str, str]:
        if self.p_field.selections:
            return {
                selection.name: selection.as_conf_string()
                for selection in self.p_field.selections.values()
            }
        else:
            return {}


def config_template(context: ConfigContext):
    return CONFIG_TEMPLATE.render(context=context)



from attrs import define


@define
class CustomScalarDefinition:
    type_name: str
    graphql_name: str
    deserialized_type: str
    property_type: str
    include_path: str


# An ISO-8601 encoded datetime.
DateTimeScalarDefinition = CustomScalarDefinition(
    type_name="qtgql::customscalars::DateTimeScalar",
    graphql_name="DateTime",
    deserialized_type="QDateTime",
    property_type="QString",
    include_path="<qtgql/customscalars/customscalars.hpp>",
)
DateScalarDefinition = CustomScalarDefinition(
    type_name="qtgql::customscalars::DateScalar",
    graphql_name="Date",
    deserialized_type="QDate",
    property_type="QString",
    include_path="<qtgql/customscalars/customscalars.hpp>",
)
TimeScalarDefinition = CustomScalarDefinition(
    type_name="qtgql::customscalars::TimeScalar",
    graphql_name="Time",
    deserialized_type="QTime",
    property_type="QString",
    include_path="<qtgql/customscalars/customscalars.hpp>",
)

DecimalScalarDefinition = CustomScalarDefinition(
    type_name="qtgql::customscalars::DecimalScalar",
    graphql_name="Decimal",
    deserialized_type="QString",
    property_type="QString",
    include_path="<qtgql/customscalars/customscalars.hpp>",
)


CustomScalarMap = dict[str, CustomScalarDefinition]
CUSTOM_SCALARS: CustomScalarMap = {
    DateTimeScalarDefinition.graphql_name: DateTimeScalarDefinition,
    DateScalarDefinition.graphql_name: DateScalarDefinition,
    TimeScalarDefinition.graphql_name: TimeScalarDefinition,
    DecimalScalarDefinition.graphql_name: DecimalScalarDefinition,
}



import re
from pathlib import Path
from typing import Any
from typing import Optional
from typing import Type

from attr import define


@define
class FileSpec:
    path: Path
    content: str

    def dump(self) -> None:
        self.path.write_text(self.content)


class AntiForwardRef:
    """
    i.e:
    Union["someString"] would return a ForwardRef, this class is a simple hack
    to just return a type contains the name.
    Also, this is a workaround for types that reference each-other,
    otherwise it would cause recursion error.
    """

    name: str
    type_map: dict

    @classmethod
    def resolve(cls) -> Any:
        return cls.type_map[cls.name]


def anti_forward_ref(name: str, type_map: dict) -> Type[AntiForwardRef]:
    return type(name, (AntiForwardRef,), {"name": name, "type_map": type_map})


def get_operation_name(query: str) -> Optional[str]:
    if match := re.search(r"(subscription|mutation|query)(.*?({|\())", query):
        return match.group(2).replace(" ", "").strip("{").strip("(")



from __future__ import annotations

import subprocess
import warnings
from functools import cached_property
from typing import List
from typing import Optional
from typing import TYPE_CHECKING
from typing import TypedDict
from typing import Union

import graphql
from graphql import language as gql_lang
from graphql import OperationType
from graphql.language import visitor
from graphql.type import definition as gql_def

from qtgqlcodegen.compiler.builtin_scalars import BuiltinScalars
from qtgqlcodegen.compiler.operation import QtGqlOperationDefinition
from qtgqlcodegen.compiler.template import cmake_template
from qtgqlcodegen.compiler.template import CmakeTemplateContext
from qtgqlcodegen.compiler.template import operation_template
from qtgqlcodegen.compiler.template import OperationTemplateContext
from qtgqlcodegen.compiler.template import schema_types_template
from qtgqlcodegen.compiler.template import SchemaTemplateContext
from qtgqlcodegen.exceptions import QtGqlException
from qtgqlcodegen.graphql_ref import is_enum_definition
from qtgqlcodegen.graphql_ref import is_input_definition
from qtgqlcodegen.graphql_ref import is_interface_definition
from qtgqlcodegen.graphql_ref import is_list_definition
from qtgqlcodegen.graphql_ref import is_named_type_node
from qtgqlcodegen.graphql_ref import is_non_null_definition
from qtgqlcodegen.graphql_ref import is_nonnull_node
from qtgqlcodegen.graphql_ref import is_object_definition
from qtgqlcodegen.graphql_ref import is_operation_def_node
from qtgqlcodegen.graphql_ref import is_scalar_definition
from qtgqlcodegen.graphql_ref import is_union_definition
from qtgqlcodegen.objecttype import EnumMap
from qtgqlcodegen.objecttype import EnumValue
from qtgqlcodegen.objecttype import GqlTypeHinter
from qtgqlcodegen.objecttype import QtGqlEnumDefinition
from qtgqlcodegen.objecttype import QtGqlFieldDefinition
from qtgqlcodegen.objecttype import QtGqlInputFieldDefinition
from qtgqlcodegen.objecttype import QtGqlInputObjectTypeDefinition
from qtgqlcodegen.objecttype import QtGqlInterfaceDefinition
from qtgqlcodegen.objecttype import QtGqlObjectTypeDefinition
from qtgqlcodegen.objecttype import QtGqlVariableDefinition
from qtgqlcodegen.utils import anti_forward_ref
from qtgqlcodegen.utils import FileSpec

if TYPE_CHECKING:  # pragma: no cover
    from qtgqlcodegen.config import QtGqlConfig

introspection_query = graphql.get_introspection_query(descriptions=True)


class GeneratedNamespace(TypedDict):
    handlers: str
    objecttypes: str


class QtGqlVisitor(visitor.Visitor):
    """Creates handlers for root operations.

    Also injects id field to types that have id field but not explicitly
    queried it.
    """

    def __init__(self, evaluator: SchemaEvaluator):
        super().__init__()
        self.operations: dict[str, QtGqlOperationDefinition] = {}
        self.evaluator = evaluator

    def _parse_variable_definition(
        self,
        var: gql_lang.VariableDefinitionNode,
    ) -> QtGqlVariableDefinition:
        return QtGqlVariableDefinition(
            name=var.variable.name.value,
            type=self.evaluator.get_type(var.type),
            type_map=self.evaluator._objecttypes_def_map,
            scalars=self.evaluator.config.custom_scalars,
            enums=self.evaluator._enums_def_map,
            default_value=var.default_value,
        )

    def enter_operation_definition(self, node, key, parent, path, ancestors) -> None:
        if operation := is_operation_def_node(node):
            if operation.operation in (
                OperationType.QUERY,
                OperationType.MUTATION,
                OperationType.SUBSCRIPTION,
            ):
                assert operation.name, "QtGql enforces operations to have names."
                operation_vars: list[QtGqlVariableDefinition] = []

                # input variables
                if variables_def := operation.variable_definitions:
                    for var in variables_def:
                        operation_vars.append(self._parse_variable_definition(var))

                operation_definition = QtGqlOperationDefinition.from_definition(
                    operation_def=operation,
                    evaluator=self.evaluator,
                    directives=node.directives,
                    variables=operation_vars,
                )
                self.operations[operation_definition.name] = operation_definition


class SchemaEvaluator:
    def __init__(self, config: QtGqlConfig):
        self.config = config
        self._objecttypes_def_map: dict[str, QtGqlObjectTypeDefinition] = {}
        self._enums_def_map: EnumMap = {}
        self._input_objects_def_map: dict[str, QtGqlInputObjectTypeDefinition] = {}
        self._interfaces_map: dict[str, QtGqlInterfaceDefinition] = {}
        self._operations: dict[str, QtGqlOperationDefinition] = {}
        self.operation_types: dict[OperationType, QtGqlObjectTypeDefinition] = {}

    def get_interface_by_name(self, name: str) -> Optional[QtGqlInterfaceDefinition]:
        return self._interfaces_map.get(name, None)

    def get_objecttype_by_name(self, name: str) -> Optional[QtGqlObjectTypeDefinition]:
        return self._objecttypes_def_map.get(name, None)

    def get_type(self, node: gql_lang.TypeNode) -> GqlTypeHinter:
        if nonnull := is_nonnull_node(node):
            return self._evaluate_field_type(
                graphql.type.GraphQLNonNull(
                    self.schema_definition.get_type(nonnull.type.name.value),  # type: ignore
                ),
            )

        if named_type := is_named_type_node(node):
            gql_concrete = self.schema_definition.get_type(named_type.name.value)
            assert gql_concrete
            return self._evaluate_field_type(gql_concrete)
        raise NotImplementedError(node, "Type is not supported as a variable ATM")

    @cached_property
    def schema_definition(self) -> graphql.GraphQLSchema:
        return graphql.build_schema(
            (self.config.graphql_dir / "schema.graphql").resolve(True).read_text(),
        )

    @cached_property
    def root_types(self) -> List[Optional[gql_def.GraphQLObjectType]]:
        return [
            self.schema_definition.get_root_type(OperationType.QUERY),
            self.schema_definition.get_root_type(OperationType.MUTATION),
            self.schema_definition.get_root_type(OperationType.SUBSCRIPTION),
        ]

    @cached_property
    def root_types_names(self) -> str:
        return " ".join([tp.name for tp in self.root_types if tp])

    def _evaluate_field_type(self, t: gql_def.GraphQLType) -> GqlTypeHinter:
        # even though every type in qtgql has a default constructor,
        # hence there is no "real" non-null values
        # we store it as optional for generating nullable checks only for what's required.
        ret = None
        is_optional = True
        if non_null := is_non_null_definition(t):
            t = non_null.of_type
            is_optional = False

        if list_def := is_list_definition(t):
            ret = GqlTypeHinter(
                type=list,
                of_type=(self._evaluate_field_type(list_def.of_type),),
                scalars=self.config.custom_scalars,
            )
        elif scalar_def := is_scalar_definition(t):
            if builtin_scalar := BuiltinScalars.by_graphql_name(scalar_def.name):
                ret = GqlTypeHinter(type=builtin_scalar, scalars=self.config.custom_scalars)
            else:
                ret = GqlTypeHinter(
                    type=self.config.custom_scalars[scalar_def.name],
                    scalars=self.config.custom_scalars,
                )
        elif enum_def := is_enum_definition(t):
            ret = GqlTypeHinter(
                type=anti_forward_ref(name=enum_def.name, type_map=self._enums_def_map),
                scalars=self.config.custom_scalars,
            )
        elif obj_def := is_object_definition(t):
            ret = GqlTypeHinter(
                type=anti_forward_ref(name=obj_def.name, type_map=self._objecttypes_def_map),
                scalars=self.config.custom_scalars,
            )
        elif union_def := is_union_definition(t):
            ret = GqlTypeHinter(
                type=Union,
                of_type=tuple(
                    GqlTypeHinter(
                        type=anti_forward_ref(
                            name=possible.name,
                            type_map=self._objecttypes_def_map,
                        ),
                        scalars=self.config.custom_scalars,
                    )
                    for possible in self.schema_definition.get_possible_types(union_def)
                ),
                scalars=self.config.custom_scalars,
            )
        elif input_def := is_input_definition(t):
            concrete = self.schema_definition.get_type(input_def.name)
            assert isinstance(concrete, gql_def.GraphQLInputObjectType)
            ret = GqlTypeHinter(
                type=self._evaluate_input_type(input_def),
                scalars=self.config.custom_scalars,
            )

        elif interface_def := is_interface_definition(t):
            ret = GqlTypeHinter(
                type=self._evaluate_interface_type(interface_def),
                of_type=(),
                scalars=self.config.custom_scalars,
            )
        if not ret:  # pragma: no cover
            raise NotImplementedError(f"type {t} not supported yet")

        if is_optional:
            return GqlTypeHinter(type=Optional, of_type=(ret,), scalars=self.config.custom_scalars)
        return ret

    def _evaluate_field(self, name: str, field: gql_def.GraphQLField) -> QtGqlFieldDefinition:
        return QtGqlFieldDefinition(
            type=self._evaluate_field_type(field.type),
            name=name,
            type_map=self._objecttypes_def_map,
            scalars=self.config.custom_scalars,
            enums=self._enums_def_map,
            description=field.description,
        )

    def _evaluate_input_field(
        self,
        name: str,
        field: gql_def.GraphQLInputField,
    ) -> QtGqlInputFieldDefinition:
        return QtGqlInputFieldDefinition(
            type=self._evaluate_field_type(field.type),
            name=name,
            type_map=self._objecttypes_def_map,
            scalars=self.config.custom_scalars,
            enums=self._enums_def_map,
            description=field.description,
        )

    def _evaluate_object_type(
        self,
        type_: gql_def.GraphQLObjectType,
    ) -> Optional[QtGqlObjectTypeDefinition]:
        t_name: str = type_.name
        if evaluated := self._objecttypes_def_map.get(t_name, None):
            return evaluated
        if type_.name not in self.root_types_names:
            try:
                id_field = type_.fields["id"]
                if nonull := is_non_null_definition(id_field.type):
                    id_scalar = graphql.type.scalars.GraphQLID
                    if nonull.of_type is not id_scalar:
                        raise QtGqlException(
                            f"id field type must be of the {id_scalar} scalar!"
                            f"\n Got: {nonull.of_type}",
                        )
                else:
                    warnings.warn(
                        stacklevel=2,
                        message="It is best practice to have id field of type ID!"
                        f"\ntype {type_} has: {id_field}",
                    )
            except KeyError:
                warnings.warn(
                    stacklevel=2,
                    message="QtGql enforces types to have ID field"
                    f"type {type_} does not not define an id field.\n"
                    f"fields: {type_.fields}",
                )
        implements = [self._evaluate_interface_type(interface) for interface in type_.interfaces]

        ret = QtGqlObjectTypeDefinition(
            name=t_name,
            implements=implements,
            docstring=type_.description,
            fields_dict={
                name: self._evaluate_field(name, field) for name, field in type_.fields.items()
            },
        )
        self._objecttypes_def_map[ret.name] = ret
        for interface in type_.interfaces:
            qtgql_interface = self._evaluate_interface_type(interface)
            qtgql_interface.implementations[type_.name] = ret
        return ret

    def _evaluate_input_type(
        self,
        type_: gql_def.GraphQLInputObjectType,
    ) -> QtGqlInputObjectTypeDefinition:
        ret = self._input_objects_def_map.get(type_.name, None)
        if not ret:
            ret = QtGqlInputObjectTypeDefinition(
                name=type_.name,
                docstring=type_.description,
                fields_dict={
                    name: self._evaluate_input_field(name, field)
                    for name, field in type_.fields.items()
                },
            )
            self._input_objects_def_map[ret.name] = ret
        return ret

    def _evaluate_interface_type(
        self,
        interface: gql_def.GraphQLInterfaceType,
    ) -> QtGqlInterfaceDefinition:
        if ret := self._interfaces_map.get(interface.name, None):
            return ret
        implements = [self._evaluate_interface_type(base) for base in interface.interfaces]
        ret = QtGqlInterfaceDefinition(
            name=interface.name,
            implements=implements,
            docstring=interface.description,
            fields_dict={
                name: self._evaluate_field(name, field) for name, field in interface.fields.items()
            },
        )
        self._interfaces_map[ret.name] = ret
        return ret

    def _evaluate_enum(self, enum: gql_def.GraphQLEnumType) -> Optional[QtGqlEnumDefinition]:
        name: str = enum.name

        if definition := self._enums_def_map.get(name, None):
            return definition

        ret = QtGqlEnumDefinition(
            name=name,
            members=[
                EnumValue(name=name, description=val.description or "")
                for name, val in enum.values.items()
            ],
        )

        self._enums_def_map[name] = ret
        return ret

    def parse_schema_concretes(self) -> None:
        for name, type_ in self.schema_definition.type_map.items():
            if name.startswith("__"):
                continue
            if object_definition := is_object_definition(type_):
                if object_type := self._evaluate_object_type(object_definition):
                    if object_definition is self.schema_definition.query_type:
                        self.operation_types[OperationType.QUERY] = object_type
                    elif object_definition is self.schema_definition.mutation_type:
                        self.operation_types[OperationType.MUTATION] = object_type
                    elif object_definition is self.schema_definition.subscription_type:
                        self.operation_types[OperationType.SUBSCRIPTION] = object_type

            elif enum_def := is_enum_definition(type_):
                if enum := self._evaluate_enum(enum_def):
                    self._enums_def_map[enum.name] = enum

    def parse_operations(self) -> None:
        with (self.config.graphql_dir / "operations.graphql").open() as f:
            operations = graphql.parse(f.read())

        # validate the operation against the static schema
        if errors := graphql.validate(self.schema_definition, operations):
            raise QtGqlException([error.formatted for error in errors])

        # get QtGql fields from the query AST and inject ID field.
        operation_miner = QtGqlVisitor(self)
        visitor.visit(operations, operation_miner)
        self._operations.update(operation_miner.operations)

    def generate(self) -> list[FileSpec]:
        self.parse_schema_concretes()
        self.parse_operations()
        context = SchemaTemplateContext(
            enums=list(self._enums_def_map.values()),
            types=[
                t
                for name, t in self._objecttypes_def_map.items()
                if name not in BuiltinScalars.keys() and name not in self.root_types_names
            ],
            interfaces=list(self._interfaces_map.values()),
            input_objects=list(self._input_objects_def_map.values()),
            config=self.config,
        )

        operations: list[FileSpec] = []
        for op_name, op in self._operations.items():
            operations.append(
                FileSpec(
                    content=operation_template(
                        OperationTemplateContext(
                            operation=op,
                            config=self.config,
                        ),
                    ),
                    path=self.config.generated_dir / f"{op_name}.hpp",
                ),
            )
        schema = FileSpec(
            content=schema_types_template(context),
            path=self.config.generated_dir / "schema.hpp",
        )

        return [schema, *operations]

    def dump(self):
        """:param file: Path to the directory the codegen would dump to."""
        headers = self.generate()

        args = ["clang-format"] + [
            str(f.path) for f in headers if f.path.suffix in (".cpp", ".h", ".hpp")
        ]
        cmake = FileSpec(
            content=cmake_template(CmakeTemplateContext(config=self.config, sources=headers)),
            path=self.config.generated_dir / "CMakeLists.txt",
        )
        headers.append(cmake)
        args.append("-i")
        for f in headers:
            f.dump()
        subprocess.run(args)



from __future__ import annotations

import contextlib
import enum
from functools import cached_property
from typing import Any
from typing import Generic
from typing import Optional
from typing import TYPE_CHECKING
from typing import TypeVar

import attrs
from attrs import define
from typingref import TypeHinter
from typingref import UNSET

from qtgqlcodegen.compiler.builtin_scalars import BuiltinScalar
from qtgqlcodegen.compiler.operation import QtGqlQueriedObjectType
from qtgqlcodegen.cppref import QtGqlTypes
from qtgqlcodegen.utils import AntiForwardRef

if TYPE_CHECKING:
    from qtgqlcodegen.runtime.custom_scalars import CustomScalarDefinition
    from qtgqlcodegen.runtime.custom_scalars import CustomScalarMap


class Kinds(enum.Enum):
    SCALAR = "SCALAR"
    OBJECT = "OBJECT"
    ENUM = "ENUM"
    LIST = "LIST"
    UNION = "UNION"
    NON_NULL = "NON_NULL"
    INTERFACE = "INTERFACE"
    INPUT_OBJECT = "INPUT_OBJECT"


@define(slots=False)
class QtGqlBaseTypedNode:
    name: str
    type: GqlTypeHinter
    type_map: dict[str, QtGqlObjectTypeDefinition]
    enums: EnumMap
    scalars: CustomScalarMap

    @cached_property
    def is_custom_scalar(self) -> Optional[CustomScalarDefinition]:
        return self.type.is_custom_scalar

    @cached_property
    def member_type(self) -> str:
        """
        :returns: Annotation of the field based on the real type,
        meaning that the private attribute would be of that type.
        this goes for init and the property setter.
        """
        return self.type.member_type


T = TypeVar("T")


@define(slots=False)
class QtGqlVariableDefinition(Generic[T], QtGqlBaseTypedNode):
    default_value: Optional[T] = UNSET

    def json_repr(self, attr_name: Optional[str] = None) -> str:
        if not attr_name:
            attr_name = self.name
        if self.type.is_input_object_type:
            raise NotImplementedError
        elif self.type.is_builtin_scalar:
            return f"{attr_name}.value()"
        elif self.type.is_enum:
            raise NotImplementedError
        elif self.is_custom_scalar:
            raise NotImplementedError

        raise NotImplementedError(f"{self.type} is not supported as an input type ATM")


@define(slots=False)
class BaseQtGqlFieldDefinition(QtGqlBaseTypedNode):
    description: Optional[str] = ""


@define(slots=False)
class QtGqlInputFieldDefinition(BaseQtGqlFieldDefinition, QtGqlVariableDefinition):
    ...


@define(slots=False)
class QtGqlFieldDefinition(BaseQtGqlFieldDefinition):
    @cached_property
    def default_value(self):
        if builtin_scalar := self.type.is_builtin_scalar:
            return builtin_scalar.default_value
        if self.type.is_object_type:
            return "{}"

        if self.type.is_model:
            # this would just generate the model without data.
            raise NotImplementedError

        if self.type.is_custom_scalar:
            return "{}"

        if enum_def := self.type.is_enum:
            raise NotImplementedError

        raise NotImplementedError

    @cached_property
    def fget_annotation(self) -> str:
        """This annotates the value that is QML-compatible."""
        if custom_scalar := self.type.is_custom_scalar:
            return TypeHinter.from_annotations(
                custom_scalar.property_type,
            ).stringify()
        if self.type.is_enum:
            return "int"

        return self.member_type

    @cached_property
    def property_type(self) -> str:
        if self.type.is_builtin_scalar or self.type.is_custom_scalar:
            return self.fget_annotation
        else:
            if self.type.is_enum:
                # QEnum value must be int
                return "int"
            # might be a model, which is also QObject
            # graphql doesn't support scalars or enums in Unions ATM.
            assert (
                self.type.is_model
                or self.type.is_object_type
                or self.type.is_interface
                or self.type.is_union
            )
            return "QObject"

    @cached_property
    def getter_name(self) -> str:
        return f"get_{self.name}"

    @cached_property
    def setter_name(self) -> str:
        return f"{self.name}_setter"

    @cached_property
    def signal_name(self) -> str:
        return f"{self.name}Changed"

    @cached_property
    def private_name(self) -> str:
        return f"m_{self.name}"

    @cached_property
    def can_select_id(self) -> Optional[QtGqlFieldDefinition]:
        object_type = self.type.is_object_type or self.type.is_interface
        if not object_type:
            if self.type.is_model:
                object_type = self.type.is_model.is_object_type
        if object_type:
            return object_type.has_id_field


@define(slots=False)
class BaseGqlTypeDefinition:
    name: str
    fields_dict: dict[str, QtGqlFieldDefinition]
    docstring: Optional[str] = ""

    @property
    def fields(self) -> list[QtGqlFieldDefinition]:
        return list(self.fields_dict.values())


@define(slots=False)
class QtGqlObjectTypeDefinition(BaseGqlTypeDefinition):
    implements: list[QtGqlInterfaceDefinition] = attrs.Factory(list)

    @cached_property
    def has_id_field(self) -> Optional[QtGqlFieldDefinition]:
        return self.fields_dict.get("id", None)

    @cached_property
    def id_is_optional(self) -> Optional[QtGqlFieldDefinition]:
        if id_f := self.has_id_field:
            if id_f.type.is_optional():
                return id_f

    def __attrs_post_init__(self):
        for base in self.implements:
            if not base.implementations.get(self.name):
                base.implementations[self.name] = self


@define(slots=False)
class QtGqlInterfaceDefinition(QtGqlObjectTypeDefinition):
    implementations: dict[str, BaseGqlTypeDefinition] = attrs.field(factory=dict)


@define(slots=False)
class QtGqlInputObjectTypeDefinition(BaseGqlTypeDefinition):
    fields_dict: dict[str, QtGqlInputFieldDefinition] = attrs.field(factory=dict)  # type: ignore


@define
class EnumValue:
    """encapsulates enumValues from introspection, maps to an Enum member."""

    name: str
    description: str = ""


@define
class QtGqlEnumDefinition:
    name: str
    members: list[EnumValue]


EnumMap = dict[str, "QtGqlEnumDefinition"]


def freeze(self, key, value):  # pragma: no cover
    raise PermissionError("setattr called on frozen type")


class GqlTypeHinter(TypeHinter):
    def __init__(
        self,
        type: Any,
        scalars: CustomScalarMap,
        of_type: tuple[GqlTypeHinter, ...] = (),
    ):
        self.type = type
        self.of_type: tuple[GqlTypeHinter, ...] = of_type
        self.scalars = scalars
        self.__setattr__ = freeze  # type: ignore

    @cached_property
    def optional_maybe(self) -> GqlTypeHinter:
        return self if not super().is_optional() else self.of_type[0]

    @cached_property
    def is_union(self) -> bool:
        return super().is_union()

    @cached_property
    def is_object_type(self) -> Optional[QtGqlObjectTypeDefinition]:
        t_self = self.optional_maybe.type
        if self.is_interface:
            return None
        with contextlib.suppress(TypeError):
            if issubclass(t_self, AntiForwardRef):
                ret = t_self.resolve()
                if isinstance(ret, QtGqlObjectTypeDefinition):
                    return ret

    @cached_property
    def is_queried_object_type(self) -> Optional[QtGqlQueriedObjectType]:
        t_self = self.optional_maybe.type
        if isinstance(t_self, QtGqlQueriedObjectType):
            return t_self

    @cached_property
    def is_input_object_type(self) -> Optional[QtGqlInputObjectTypeDefinition]:
        t_self = self.optional_maybe.type
        if isinstance(t_self, QtGqlInputObjectTypeDefinition):
            return t_self

    @cached_property
    def is_model(self) -> Optional[GqlTypeHinter]:
        t_self = self.optional_maybe
        if t_self.is_list():
            # scalars or unions are not supported in lists yet (valid graphql spec though)
            return t_self.of_type[0]

    @cached_property
    def is_interface(self) -> Optional[QtGqlInterfaceDefinition]:
        t_self = self.optional_maybe.type
        if isinstance(t_self, QtGqlInterfaceDefinition):
            return t_self

    @cached_property
    def is_enum(self) -> Optional[QtGqlEnumDefinition]:
        t_self = self.optional_maybe.type
        with contextlib.suppress(TypeError):
            if issubclass(t_self, AntiForwardRef):
                if isinstance(t_self.resolve(), QtGqlEnumDefinition):
                    return t_self.resolve()

    @cached_property
    def is_builtin_scalar(self) -> Optional[BuiltinScalar]:
        t_self = self.optional_maybe.type
        if isinstance(t_self, BuiltinScalar):
            return t_self

    @cached_property
    def is_custom_scalar(self) -> Optional[CustomScalarDefinition]:
        t_self = self.optional_maybe.type
        if t_self in self.scalars.values():
            return t_self

    @cached_property
    def member_type(self) -> str:
        """
        :returns: Annotation of the field based on the real type,
        meaning that the private attribute would be of that type.
        this goes for init and the property setter. They are optional by default,
        (at the template) so unwrap optional first
        """
        t_self = self.optional_maybe

        # int, str, float etc...
        if builtin_scalar := t_self.is_builtin_scalar:
            return builtin_scalar.tp

        if scalar := t_self.is_custom_scalar:
            return scalar.type_name
        if gql_enum := t_self.is_enum:
            return gql_enum.name
        if model_of := t_self.is_model:
            return f"{QtGqlTypes.QGraphQLList.name}[{model_of.member_type}]"
        if object_def := t_self.is_object_type or t_self.is_interface:
            return f"std::shared_ptr<{object_def.name}>"
        if q_object_def := t_self.is_queried_object_type:
            return f"{q_object_def.name}"
        if t_self.is_union:
            return "std::variant<" + ", ".join(th.member_type for th in t_self.of_type) + ">"
        if input_obj := t_self.is_input_object_type:
            return input_obj.name

        raise NotImplementedError  # pragma no cover

    def as_annotation(self, object_map=None):  # pragma: no cover
        raise NotImplementedError("not safe to call on this type")



from __future__ import annotations

import glob
import importlib
import os
import sys
from pathlib import Path
from typing import TYPE_CHECKING

import rich
import typer

if TYPE_CHECKING:
    from qtgqlcodegen.config import QtGqlConfig

console = rich.console.Console()
app = typer.Typer(pretty_exceptions_show_locals=False)

QTGQL_CONFIG_FNAME = "qtgqlconfig.py"


def _get_config() -> QtGqlConfig:
    console.print("[bold blue]Looking for you config file...")
    res = glob.glob(f"{QTGQL_CONFIG_FNAME}")

    if not len(res) > 1:
        console.print(
            f"[bold red]Found more than one config file. {len(res)}"
            f" found: {os.linesep.join(res)}",
        )
        typer.Abort()
    if len(res) == 0:
        console.print("[bold red]Found more than one config file using the first one")
        typer.Abort()
    mod_path = Path(res[0]).resolve(True)
    spec = importlib.util.spec_from_file_location(QTGQL_CONFIG_FNAME, mod_path)
    assert spec
    module = importlib.util.module_from_spec(spec)
    sys.modules[QTGQL_CONFIG_FNAME] = module
    assert spec.loader
    spec.loader.exec_module(module)
    return module.config


@app.command()
def gen():
    """Generates types based on your `QtGqlConfig` configuration object."""
    console.print("[bold blue]Generating...")
    with console.status("Still generating...") as s:
        config = _get_config()
        s.update("[green]Configuration file loaded")
        s.update("[bold blue]Just a second I need some coffee ☕")
        config.generate()

    console.print(
        "[bold green]Types were generated to"
        f"[link={config.generated_dir.resolve()}]"
        f"file://{config.generated_dir.resolve()}[/link] successfully!",
    )


@app.command()
def hotreload():  # pragma: no cover
    raise NotImplementedError


def entrypoint():  # pragma: no cover
    app()
